/*
 Ardunio Uno 3 Intervalometer "Pro-Timer Free" with Tiny I2C - RTC Time Shield

  Pro-Timer Free
  Gunther Wegner
  http://gwegner.de
  http://lrtimelapse.com


 Ardunio Uno 3 Intervalometer "Pro-Timer Free" with Tiny I2C - RTC Time Shield
 
  www.LRE-Photography.com
  LRE Photography - Timelapse

  Version: 0.3
  Hardware: Arduino Uno 3, SainSmarth LCD KeyPad Shield, Tiny RTC, with and without SainSmarth 2 Relais Module

  Without Relais Module:

  digitalWrite(CAPTURE, HIGH) changed to digitalWrite(CAPTURE, LOW)
  and
  digitalWrite(CAPTURE, LOW) change to digitalWrite(CAPTURE, HIGH)


  Lines wich must be changed 116, 646 und 650

*/


#include <LiquidCrystal.h>
#include <RTClib.h>
#include <Wire.h>
#include "LCD_Keypad_Reader.h"      // credits to: http://www.hellonull.com/?p=282

// #include <EEPROM.h>
// EEPROM.read(ADDRESS);
// EEPROM.write(ADDRESS, DATA);

RTC_DS1307 RTC;

LCD_Keypad_Reader keypad;
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);  //Pin assignments for SainSmart LCD Keypad Shield

const int CAPTURE = 12;
const int POWER = 11;

// KEYPAD
const int NONE = 0;           // Key constants
const int SELECT = 1;
const int LEFT = 2;
const int UP = 3;
const int DOWN = 4;
const int RIGHT = 5;

const int keyRepeatRate = 100;      // when held, key repeats 1000 / keyRepeatRate times per second
const int keySampleRate = 50;      // ms between checking keypad for key // original: 50

int localKey = 0;           // The current pressed key
int lastKeyPressed = -1;        // The last pressed key

unsigned long lastKeyCheckTime = 0;
unsigned long lastKeyPressTime = 0;

// START, END, INTERVAL
uint8_t startHour = 0;
uint8_t startMinute = 00;
uint8_t startSecond = 01;

uint8_t endHour = 18;
uint8_t endMinute = 00;
uint8_t endSecond = 0;

uint8_t intervalSecond = 04; // 0-59
uint8_t intervalMinute = 0; // 0-59
uint8_t intervalHour = 0; // 0-23
long interval = 60; // in seconds


long captures = 0; // per day
long lastCapture = 0; // in millis
boolean doFirstCapture = false;

// MENU
const int MENU_MAIN = 0;
const int MENU_START = 1;
const int MENU_END = 2;
const int MENU_RUN = 3;
const int MENU_INTERVAL = 4;

int currentMenu = 0;
int menuMainCurrent = 0;
int posCursor = 0; // 0: off, 1: first, 2: second, 3: third

// BACKLIGHT
const int BACK_LIGHT = 10;
boolean isBacklight = false;
unsigned long backlightTime = 0;
unsigned long backlightDuration = 120000; //  30s

// CAPTURE
unsigned long delayCapture = 500;

// RUNNING
boolean isRunning = false;

// SETMODE
boolean isSetmode = false;

// POWER
boolean doPower = false;
unsigned long delayPower = 2000;

// SETUP

void setup() {
  pinMode(CAPTURE, OUTPUT);
  digitalWrite(CAPTURE, HIGH);

  if ( doPower ) {
    pinMode(POWER, OUTPUT);
    digitalWrite(POWER, HIGH);
  }

  Serial.println("Setup");

  Serial.begin(9600);
  Wire.begin();
  RTC.begin();

  if (! RTC.isrunning()) {
    Serial.println("RTC is NOT running!");
    // following line sets the RTC to the date & time this sketch was compiled
    RTC.adjust(DateTime(__DATE__, __TIME__));
  }
  else {
    Serial.println("RTC is running!");
  }

  pinMode(BACK_LIGHT, OUTPUT);
  digitalWrite(BACK_LIGHT, HIGH);
  isBacklight = true;
  backlightTime = millis();

  lcd.begin(16, 2);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Time Lapse");
  lcd.setCursor(0, 1);
  lcd.print("Version 0.3");

  lastKeyCheckTime = millis();

  delay(1000);

}



// LOOP

void loop() {

  if (  millis() > (backlightTime + backlightDuration) ) {
    if (isBacklight) digitalWrite(BACK_LIGHT, LOW);
    isBacklight = false ;

  }
  else {
    if (!isBacklight) digitalWrite(BACK_LIGHT, HIGH);
    isBacklight = true ;
  }


  if (millis() > lastKeyCheckTime + keySampleRate) {
    lastKeyCheckTime = millis();
    localKey = keypad.getKey();
    //Serial.println( localKey );
    //Serial.println( lastKeyPressed );

    if (localKey != lastKeyPressed) {
      processKey();
    } else {
      // key value has not changed, key is being held down, has it been long enough?
      // (but don't process localKey = 0 = no key pressed)
      if (localKey != 0 && millis() > lastKeyPressTime + keyRepeatRate) {
        // yes, repeat this key
        if ( (localKey == UP ) || ( localKey == DOWN ) ) {
          processKey();
        }
      }
    }

    lcd.clear();
    printScreen();
  }


  if (isRunning && (interval > 0) ) {
    DateTime now = RTC.now();
    DateTime start ( now.year(), now.month(), now.day(), startHour, startMinute, startSecond);
    DateTime end ( now.year(), now.month(), now.day(), endHour, endMinute, endSecond);

    //showDate("start:", start);

    TimeSpan sinceStart = now - start;
    TimeSpan toEnd = end - now;

    // showTimeSpan("sinceStart", sinceStart);
    // showTimeSpan("toEnd", toEnd);

    // CHECK CAPTURE
    if ( (sinceStart.totalseconds() > 0) && (toEnd.totalseconds() > 0) ) {

      //Serial.print(millis());
      //Serial.print(":");
      //Serial.println(lastCapture);

      if (  millis() > (lastCapture + (interval * 1000)) || doFirstCapture ) {
        lastCapture = millis();
        doFirstCapture = false;
        Serial.print("CAPTURE ..");
        capture();
        Serial.print(" done");
        Serial.println();
      }
      else {
        long rest = lastCapture + (interval * 1000) - millis();
        //Serial.println(rest);

        static char str[16];
        sprintf(str, "%05d", int( (float)rest / 1000.0 ) + 1 );

        lcd.setCursor(0, 1);
        lcd.print(str);


      }
    }

  }


  delay(100);
}


// KEYPAD

void processKey() {

  if (  lastKeyPressed != localKey ) {
    backlightTime = millis();
  }

  if (isBacklight) {


    lastKeyPressed = localKey;
    lastKeyPressTime = millis();

    switch ( currentMenu ) {

      case MENU_MAIN:
        if ( localKey == DOWN && menuMainCurrent == 0 ) {
          menuMainCurrent = 1;
          lcd.clear();
        }
        if ( localKey == UP && menuMainCurrent == 1 ) {
          menuMainCurrent = 0;
          lcd.clear();
        }
        if ( localKey == RIGHT ) {

          if ( menuMainCurrent == 0 ) {
            currentMenu = MENU_START;
          }

          if ( menuMainCurrent == 1 ) {
            currentMenu = MENU_RUN;
            isRunning = true;
            doFirstCapture = true;
          }

          posCursor = 0;
        }

        if ( localKey == LEFT ) {

        }
        break;




      case MENU_START:
        if ( localKey == UP ) {
          startHour = setTimeUpHours(1, 2, startHour);
          startMinute = setTimeUpMinutes(3, 4, startMinute);
        }

        if ( localKey == DOWN ) {
          startHour = setTimeDown(1, 2, startHour);
          startMinute = setTimeDown(3, 4, startMinute);
        }

        if ( localKey == RIGHT ) {

          if ( isSetmode) {
            if ( posCursor == 4 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor++;
            }
          }
          else {
            currentMenu = MENU_END;
            posCursor = 0;
            setCursor();
          }


        }

        if ( localKey == LEFT ) {
          if ( isSetmode) {
            if ( posCursor == 1 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor--;
            }
          }
          else {
            currentMenu = MENU_MAIN ;
            posCursor = 0;
            setCursor();
            isRunning = false;
          }
        }

        if ( localKey == SELECT ) {
          if (isSetmode) {
            isSetmode = false;
            posCursor = 0;
          }
          else {
            isSetmode = true;
            posCursor = 1;
          }

        }


        break;

      case MENU_END:
        if ( localKey == UP ) {
          endHour = setTimeUpHours(1, 2, endHour);
          endMinute = setTimeUpMinutes(3, 4, endMinute);
        }

        if ( localKey == DOWN ) {
          endHour = setTimeDown(1, 2, endHour);
          endMinute = setTimeDown(3, 4, endMinute);
        }

        if ( localKey == RIGHT ) {
          if ( isSetmode) {
            if ( posCursor == 4 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor++;
            }
          }
          else {
            currentMenu = MENU_INTERVAL;
            posCursor = 0;
            setCursor();
          }



        }

        if ( localKey == LEFT ) {
          if ( isSetmode) {
            if ( posCursor == 1 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor--;
            }
          }
          else {
            currentMenu = MENU_START;
            posCursor = 0;
            setCursor();
          }


        }

        if ( localKey == SELECT ) {
          if (isSetmode) {
            isSetmode = false;
            posCursor = 0;
          }
          else {
            isSetmode = true;
            posCursor = 1;
          }
        }

        break;

      case MENU_INTERVAL:
        if ( localKey == UP ) {
          intervalHour = setTimeUpHours(1, 2, intervalHour);
          intervalMinute = setTimeUpMinutes(3, 4, intervalMinute);
          intervalSecond = setTimeUpSeconds(5, 6, intervalSecond);
        }

        if ( localKey == DOWN ) {
          intervalHour = setTimeDown(1, 2, intervalHour);
          intervalMinute = setTimeDown(3, 4, intervalMinute);
          intervalSecond = setTimeDown(5, 6, intervalSecond);
        }

        interval = intervalHour * 3600 + intervalMinute * 60 + intervalSecond;

        if ( localKey == RIGHT ) {

          if ( isSetmode) {
            if ( posCursor == 6 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor++;
            }
          }
          else {
            currentMenu = MENU_MAIN;
            posCursor = 0;
            setCursor();
          }


        }

        if ( localKey == LEFT ) {

          if ( isSetmode) {
            if ( posCursor == 1 )  {
              posCursor = 0;
              isSetmode = false;
            }
            else {
              posCursor--;
            }
          }
          else {
            currentMenu = MENU_END;
            posCursor = 0;
            setCursor();
          }


        }

        if ( localKey == SELECT ) {
          if (isSetmode) {
            isSetmode = false;
            posCursor = 0;
          }
          else {
            isSetmode = true;
            posCursor = 1;
          }
        }

        break;

      case MENU_RUN:
        if ( localKey == UP ) {

        }

        if ( localKey == DOWN ) {

        }

        if ( localKey == RIGHT ) {
          currentMenu = MENU_MAIN;
          isRunning = false;
        }

        if ( localKey == LEFT ) {
          currentMenu = MENU_MAIN;
          isRunning = false;
        }
        break;

    }

  }

}


// SCREEN

void printScreen() {

  switch ( currentMenu ) {

    case MENU_MAIN:
      lcd.setCursor(0, 0);
      if ( menuMainCurrent == 0 ) lcd.print("> ");
      lcd.print("Settings");

      lcd.setCursor(0, 1);
      if ( menuMainCurrent == 1 ) lcd.print("> ");
      lcd.print("Start");
      break;

    case MENU_START:
      lcd.setCursor(0, 0);
      lcd.print("START TIME H:M");
      lcd.setCursor(0, 1);

      lcd.print(  formatTimeHM(startHour, startMinute) );

      setCursor();

      break;

    case MENU_END:
      lcd.setCursor(0, 0);
      lcd.print("END TIME H:M");
      lcd.setCursor(0, 1);
      lcd.print(  formatTimeHM(endHour, endMinute) );
      setCursor();
      break;

    case MENU_INTERVAL:
      lcd.setCursor(0, 0);
      lcd.print("INTERVAL H:M:S");
      lcd.setCursor(0, 1);
      lcd.print(  formatTimeHMS(intervalHour, intervalMinute, intervalSecond) );
      setCursor();
      break;

    case MENU_RUN:
      lcd.setCursor(0, 0);
      lcd.print("RUNNING ...");
      break;



  }

}


// SET TIME

int setTimeDown(int a, int b, int t) {
  String strT = String(t);
  if ( posCursor == a &&  t >= 10 )  t -= 10;
  if ( posCursor == b &&  !strT.endsWith("0") )  t--;
  return t;
}

int setTimeUpHours(int a, int b, int t) {
  String strT = String(t);
  if ( posCursor == a &&  t <= 13 )  t += 10;
  if ( posCursor == b &&  t <= 22 && !strT.endsWith("9")  )  t++;
  return t;
}

int setTimeUpMinutes(int a, int b, int t) {
  String strT = String(t);
  if ( posCursor == a &&  t <= 49 )  t += 10;
  if ( posCursor == b &&  t <= 58 && !strT.endsWith("9") )  t++;
  return t;
}

int setTimeUpSeconds(int a, int b, int t) {
  String strT = String(t);
  if ( posCursor == a &&  t <= 49 )  t += 10;
  if ( posCursor == b &&  t <= 58 && !strT.endsWith("9") )  t++;
  return t;
}


// SET BLINKING CURSOR

void setCursor() {
  if ( posCursor == 0 ) {
    lcd.noBlink();
  }

  if ( posCursor == 1 ) {
    lcd.setCursor(0, 1);
    lcd.blink();
  }
  if ( posCursor == 2 ) {
    lcd.setCursor(1, 1);
    lcd.blink();
  }
  if ( posCursor == 3 ) {
    lcd.setCursor(3, 1);
    lcd.blink();
  }
  if ( posCursor == 4 ) {
    lcd.setCursor(4, 1);
    lcd.blink();
  }
  if ( posCursor == 5 ) {
    lcd.setCursor(6, 1);
    lcd.blink();
  }
  if ( posCursor == 6 ) {
    lcd.setCursor(7, 1);
    lcd.blink();
  }

}


// CAPTURE

void capture() {

  if (doPower) {
    digitalWrite(POWER, LOW);
    delay(delayPower);
  }

  digitalWrite(CAPTURE, LOW);
  lcd.setCursor(0, 1);
  lcd.print("*** CAPTURE ***");
  delay(delayCapture);
  digitalWrite(CAPTURE, HIGH);
  lcd.setCursor(0, 1);
  lcd.print("               ");

  if (doPower) {
    digitalWrite(POWER, HIGH);
  }

}


// FORMAT

String formatTimeHM(int a, int b) {
  static char str[5];
  sprintf(str, "%02d:%02d", a, b);
  return str;
}
String formatTimeHMS(int a, int b, int c) {
  static char str[5];
  sprintf(str, "%02d:%02d:%02d", a, b, c);
  return str;
}


// DEBUG

void showDate(const char* txt, const DateTime& dt) {
  Serial.print(txt);
  Serial.print(' ');
  Serial.print(dt.year(), DEC);
  Serial.print('/');
  Serial.print(dt.month(), DEC);
  Serial.print('/');
  Serial.print(dt.day(), DEC);
  Serial.print(' ');
  Serial.print(dt.hour(), DEC);
  Serial.print(':');
  Serial.print(dt.minute(), DEC);
  Serial.print(':');
  Serial.print(dt.second(), DEC);

  Serial.print(" = ");
  Serial.print(dt.unixtime());
  Serial.print("s / ");
  Serial.print(dt.unixtime() / 86400L);
  Serial.print("d since 1970");

  Serial.println();
}

void showTimeSpan(const char* txt, const TimeSpan& ts) {
  Serial.print(txt);
  Serial.print(" ");
  Serial.print(ts.days(), DEC);
  Serial.print(" days ");
  Serial.print(ts.hours(), DEC);
  Serial.print(" hours ");
  Serial.print(ts.minutes(), DEC);
  Serial.print(" minutes ");
  Serial.print(ts.seconds(), DEC);
  Serial.print(" seconds (");
  Serial.print(ts.totalseconds(), DEC);
  Serial.print(" total seconds)");
  Serial.println();
}


